#!/usr/bin/ruby
#



current_dir = File.dirname(File.expand_path(__FILE__))
lib_path = File.join(current_dir, '..', 'lib')
puts lib_path
$LOAD_PATH.unshift lib_path

require 'yaml'
require 'socket'
require 'devise-proxy'
require 'ruby-debug'
require 'ipaddr'

config_file = File.join(current_dir, '..', 'config', 'devise-proxy.yml')

if File.exists? config_file
	CONF = YAML::load(File.open(config_file))
	auth_host = CONF['authentication']['hostname']
	auth_port = CONF['authentication']['port']
	AUTHENTICATOR = DeviseProxy::Authenticator.new(auth_host, auth_port)
else
	puts "\nPlease create a config file prior to starting the proxy using the provided .sample file, at:\n\n\t#{config_file}\n\n"
	exit 1
end


# Subset from the all-knowing Wikipedia:
#
# 	http://en.wikipedia.org/wiki/SOCKS
#
# The authentication methods supported are numbered as follows:
# 0x00: No authentication
# 0x01: GSSAPI[10]
# 0x02: Username/Password[11]
# 0x03-0x7F: methods assigned by IANA[12]
# 0x80-0xFE: methods reserved for private use

@@SOCKS4_VERSION = 0x04

@@SOCKS5_VERSION = 0x05
@@SOCKS5_NO_AUTH = 0x00
@@SOCKS5_USERNAME_PASSWORD = 0x02
@@SOCKS5_NO_ACCEPTABLE_METHODS = 0xFF
@@SOCKS5_AUTH_VERSION = 0x01 # "must be 0x01"
@@SOCKS5_AUTH_SUCCESS = 0x00
@@SOCKS5_AUTH_FAILURE = 0x01 # "any other value = failure, connection must be closed"

# Client-requested modes.
@@SOCKS5_TCP_STREAM_CONNECTION = 0x01
@@SOCKS5_PORT_BINDING = 0x02
@@SOCKS5_UDP_PORT = 0x03

@@SOCKS5_ADDRESS_IPV4 = 0x01
@@SOCKS5_ADDRESS_DOMAIN = 0x03
@@SOCKS5_ADDRESS_IPV6 = 0x04


def log(s)
	puts "#{Time.now}: '#{s}'"
end

def validate_socks5(client)
	data = client.recv(1).bytes.to_a
	# puts "Socks v#{data[0]}"
	data[0] == @@SOCKS5_VERSION
end

def authenticate(username, password)
	# debugger
	result = AUTHENTICATOR.authenticate(username, password)
	log("Authentication results: #{result}.")
	return result[0] == '201'
end

def not_supported(msg, client)
	log(msg)
 	client.close
end

def forward_traffic(client)
	sock = TCPSocket.new(CONF['forwarding']['hostname'], CONF['forwarding']['port'])
	req = client.read
	sock.write(req)
	resp = sock.read
	puts(resp)
	client.write(resp)
end

def read_port(client)
	port_a = client.recv(2).bytes.to_a
	port = (port_a[0] << 8) + port_a[1]
	port
end

def handle_socks5(client)
	log("Requesting credentials from client.")
	# debugger
	challenge_bytes = [@@SOCKS5_VERSION, @@SOCKS5_USERNAME_PASSWORD].pack('C*')
	log(challenge_bytes.unpack('C*'));
	client.write(challenge_bytes)
	ver = client.recv(1).bytes.to_a[0]
	puts ver
	if ver != @@SOCKS5_AUTH_VERSION
		log("Client speaking non-sense.")
	else
		username_length = client.recv(1).bytes.to_a[0]
		username = client.recv(username_length).bytes.to_a.pack('C*')
		password_length = client.recv(1).bytes.to_a[0]
		password = client.recv(password_length).bytes.to_a.pack('C*')
		log("Attempting to authenticate #{username}:#{password}.")
		success = authenticate(username, password)
		if(success)

			client.write [@@SOCKS5_AUTH_VERSION, @@SOCKS5_AUTH_SUCCESS].pack('C*')
			# if validate_socks5(client)
			client.recv(1)
			if true
				conn_type = client.recv(1).bytes.to_a[0]
				log("Client requested type: #{conn_type}")
				# case conn_type
				# when @@SOCKS5_TCP_STREAM_CONNECTION
				# when @@SOCKS5_PORT_BINDING
				 	client.recv(1) # Eat a byte... "reserved, must be 0x00"
				 	case client.recv(1).bytes.to_a[0]
				 	when @@SOCKS5_ADDRESS_IPV4
				 		log('IPv4 mode.')
				 		ip = client.recv(4).bytes.to_a
				 		port = read_port(client)
				 		log("Client requested: #{ip}:#{port}")
				 		resp = [@@SOCKS5_VERSION, 0, 0x00, @@SOCKS5_ADDRESS_IPV4] + ip + port_a
				 		client.write resp.pack('C*')
				 		forward_traffic(client)
				 	when @@SOCKS5_ADDRESS_IPV6
				 		not_supported("IPv6 not supported.", client)
				 	when @@SOCKS5_ADDRESS_DOMAIN
				 		log("Domain mode.")
				 		domain_length = client.recv(1).bytes.to_a[0]
						domain = client.recv(domain_length).bytes.to_a.pack('C*')
						port = read_port(client)
						log("Client requested: #{domain}:#{port}")
				 		resp = [@@SOCKS5_VERSION, 0, 0x00, @@SOCKS5_ADDRESS_DOMAIN] + domain + port_a
				 		client.write resp.pack('C*')
				 		forward_traffic(client)
				 	else
				 		not_supported("Unsupported client connection type.", client)
				 	end
				# else
				 	# not_supported("Client request mode not supported, sorry!", client);
				# end 
			else
				not_supported("Invalid SOCKS version.", client)
			end
		else
			client.write [@@SOCKS5_VERSION, @@SOCKS5_AUTH_FAILURE].pack('C*')
			client.close
		end
	end
	log("Done with client connection!");
end

PROXY_PORT = 6969
server = TCPServer.new PROXY_PORT
log("Server started on port #{PROXY_PORT}.")
loop do
  Thread.start(server.accept) do |client|
    neg_string = client.recv(2).bytes.to_a
    log("Client connection received.")
    log(neg_string)
    case neg_string[0]
    when @@SOCKS4_VERSION
    	log("SOCKS v4 isn't supported, sorry!")
    when @@SOCKS5_VERSION
    	num_methods = neg_string[1]
    	log("Socks v5 detected with #{num_methods} potential authentication modes.")
    	methods = client.recv(num_methods).bytes.to_a
    	supported = false
    	methods.each do |m|
    		log("Potential mode: #{m}")
    		supported = true if m == @@SOCKS5_USERNAME_PASSWORD
    	end
    	if supported
    		log("Username/Password authentication is supported!")
    		handle_socks5(client)
    	else
    		client.puts(@@SOCKS5_VERSION)
    		client.puts(@@SOCKS5_NO_ACCEPTABLE_METHODS)
    	end
    else
    	log("Unknown/Unsupported client negotiation process.")
    end
    # (neg_string[0] == 5)
    client.close
  end
end

